@@ -27,8 +27,9 @@ async def main():
             elif row < 0:
                 assert row == -9
                 break
-    # should be under 10 seconds
-    for i in tqdm.as_completed(list(map(asyncio.sleep, [1] * 10)),
+    # should be ~1sec rather than ~50s due to async scheduling
+    for i in tqdm.as_completed([asyncio.sleep(0.01 * i)
+                                for i in range(100, 0, -1)],
                                desc="as_completed"):
         await i
 
